/* $Id: timer.c,v 1.11 2005/06/04 21:50:12 lacos Exp $ */

#define _XOPEN_SOURCE
#define _XOPEN_SOURCE_EXTENDED 1 /* enable X/Open UNIX Extension */

#include <float.h>  /* FLT_RADIX */

#include "timer.h"  /* interface */


int check_time_t(void)
{
  /*
    (1.0 / FLT_RADIX) can be represented exactly by all floating point types.
    Its value is in (0, 1).
  */
  return (time_t)(1.0 / FLT_RADIX) == (time_t)0 ? 0 : -1;
}


static struct timeval c_prev; /* C[I-1], I >= 0 */
static ldiv_t s_prev;         /* S[I-1], I >= 0 */
static long k;                /* K */


static void timer_init1(const struct timeval *current_time)
{
  /* C[-1] := current_time */
  c_prev = *current_time;

  /* S[-1] := 0 */
  s_prev.quot = 0L;
  s_prev.rem = 0L;

  /* I := 0 */
}


void timer_init(const struct timeval *current_time, long bps)
{
  timer_init1(current_time);
  /* Algorithm initialization done. */

  k = bps;
}


int timer_adv(const struct timeval *c, const ldiv_t *l, int restart)
{
  long cdiff_quot;
  ldiv_t s;
  int ret = 0;

  /*
    Already done by the caller:

    L[I] := exec_event(I)
    C[I] := current_time
  */

  if (restart)
    timer_init1(c);

  /*
    S[I] := S[I-1] + (L[I] - (C[I] - C[I-1]))

    For all I >= 0, let CDIFF[I] := C[I] - C[I-1]. C[I] and C[I-1] are integral
    multiples of one microsecond, so is CDIFF[I].
  */

  {
    struct timeval cdiff;

    /*
      Wrt. these two steps, see the caller's responsibilities before calling
      this function. 0 <= CDIFF[I] <= 1800s, I >= 0.
    */
    tv_sub(&cdiff, c, &c_prev);
    cdiff_quot = (long)(cdiff.tv_sec) * 1000000L + cdiff.tv_usec;
  }

  /*
    Theorem. For all I >= 0, S[I-1] <= CDIFF[I].
    Proof.
      If
        I = 0 (first call or restarted),
      then
        proof is 0 <= C[0] - C[-1], which is ensured by the caller (with
        checking or with restart);
      else
        I >= 1, previous sleep was uninterrupted, system time was not reset,
        thus the difference of the two "gettimeofday()"'s (= CDIFF[I]) is no
        smaller than the timeout argument of the previous "select()" (=
        S[I-1]).

        "Select()" can timeout earlier than it is told to, but only if the
        specified timeout is greater than an implementation-dependent value of
        at least 31 days. In this case, "select()" sleeps for this
        implementation-dependent value of at least 31 days, which would make
        CDIFF[I] > 1800s, and the caller wouldn't call this function without
        restart.

    Now it is known that for all I >= 0,

                S[I-1]
      ( 1)   <= CDIFF[I]
      ( 2)   <= 1.8 * 10^9

                0
      ( 3)   <= CDIFF[I]

                0
      ( 4)   <  L[I]


    If

                0
      ( 5)   <= S[I-1]

    1 then

                0
      ( 5)   <= S[I-1]
      ( 1)   <= CDIFF[I]

      Subtract CDIFF[I] (implementation: no sub-microsecond part, can be done
      exactly using only quot).

                -1.8 * 10^9
      ( 6)   <= -CDIFF[I]                      using ( 2)
             <= S[I-1] - CDIFF[I]
             <= 0

      Add L[I] (implementation: can be done with two additions, first quot,
      then sub-microsecond, at least one is positive, the other is
      non-negative).

                -1.8 * 10^9
      ( 7)   <  -1.8 * 10^9 + L[I]             using ( 4)
             <= -CDIFF[I] + L[I]
             <= (S[I-1] - CDIFF[I]) + L[I]
             <= L[I]

    2 else

                S[I-1]
      ( 8)   <  0                              using ( 5)


      Two things to note:

        - if S[I-1] is negative (I >= 0), sleeping behaves as if S[I-1] was 0,
        thus the theorem above still holds;

        - if S[I-1] is negative (I >= 0), then it is _not_ compatible with
        "ldiv()" (meaning that "ldiv()" would never return such a value),
        because the remainder is still non-negative. Anyway, this is no problem
        at all because of the way the elements of S are used.

                -1.8 * 10^9
      ( 6)   <= -CDIFF[I]
             <= 0                              using ( 3)

      Add L[I] (implementation: can be done with one addition and one copy, at
      least one of them is positive, the other one is non-negative).

                -1.8 * 10^9
      ( 7)   <  -1.8 * 10^9 + L[I]
             <= -CDIFF[I] + L[I]
      ( 9)   <= L[I]

      If

                  0
        (10)   <= -CDIFF[I] + L[I]

      2.1 then

        Add S[I-1] (implementation: tricky, all steps must be made towards
        negative infinity).

                  S[I-1]
               <= (-CDIFF[I] + L[I]) + S[I-1]
               <  -CDIFF[I] + L[I]             using ( 8)
               <= L[I]                         using ( 9)

      2.2 else

                  -1.8 * 10^9
               <  -CDIFF[I] + L[I]             using (7)
               <  0                            using (10)


        Negate (implementation: tricky).

                  0
               <  -(-CDIFF[I] + L[I])
               <  1.8 * 10^9


        Add -(1.8 * 10^9) (implementation: can be done using exactly quot).

                  -1.8 * 10^9
               <  -(-CDIFF[I] + L[I]) - 1.8 * 10^9
               <  0

        If

                    -(-CDIFF[I] + L[I]) - 1.8 * 10^9
                 <= S[I-1]

        2.2.1 then

                    -(-CDIFF[I] + L[I]) - 1.8 * 10^9
                 <= S[I-1]
          ( 8)   <  0


          Add (-CDIFF[I] + L[I]) (implementation: tricky, reuse 2.1).

                    -1.8 * 10^9
                 <= (-CDIFF[I] + L[I]) + S[I-1]
                 <  -CDIFF[I] + L[I]
          (10)   <  0

          (Accumulate drift while possible.)

        2.2.2 else

          Don't change the timer.
  */

  if (0L <= s_prev.quot) {
    s.quot = (s_prev.quot - cdiff_quot) + l->quot;
    /*
      1 <= K <= LONG_MAX/2 + 1
      0 <= *.rem < K,
      0 <= *.rem <= K - 1 <= LONG_MAX/2
    */
    s.rem = s_prev.rem + l->rem;
    if (s.rem >= k) {
      s.rem -= k;
      ++s.quot;
    }
  }
  else {
    int add;

    s.quot = -cdiff_quot + l->quot;
    s.rem = l->rem;

    if (0L <= s.quot)
      add = 1;
    else {
      ldiv_t tmp;

      tmp.quot = -(s.quot + 1L);
      if (s.rem)
        tmp.rem = k - s.rem;
      else {
        tmp.rem = 0L;
        ++tmp.quot;
      }

      tmp.quot -= (MAXCDIFFL * 1000000L);

      add = tmp.quot < s_prev.quot
          || (tmp.quot == s_prev.quot && tmp.rem <= s_prev.rem);
    }

    if (add) {
      /* "S_prev.quot + 1" is non-positive. */
      s.quot += (s_prev.quot + 1L);
      s.rem += s_prev.rem;
      if (s.rem >= k)
        s.rem -= k;
      else
        --s.quot;
    }
    else
      /* Don't change the timer. */
      s = s_prev;
  }

  /* exec_sleep(S[I]) */
  if (0L < s.quot || (0L == s.quot && 0L < s.rem)) {
    struct timeval sleep_tv;
    ldiv_t secs_usecs = ldiv(s.quot, 1000000L);

    /*
      If S[I] is not an integral multiple of 1 microsecond, round it up
      temporarily, for the purpose of sleeping. This is safe because S[I] is
      not greater than L[I], and L[I] is smaller than 31 days (I >= 0).
    */
    sleep_tv.tv_sec = secs_usecs.quot;
    sleep_tv.tv_usec = secs_usecs.rem;
    if (s.rem && 1000000L == ++sleep_tv.tv_usec) {
      sleep_tv.tv_usec = 0L;
      ++sleep_tv.tv_sec;
    }

    ret = select(0, 0, 0, 0, &sleep_tv);
  }

  /* I := I + 1 */
  c_prev = *c;
  s_prev = s;

  return ret;
}


int tv_sub(struct timeval *result, const struct timeval *tv1,
    const struct timeval *tv2)
{
  int ret = -1;
  struct timeval tmp;

  /*
    "Tv1->tv_sec" and "tv2->tv_sec" are valid instances of the same integral
    type, they can be compared.
  */
  if (tv1->tv_sec >= tv2->tv_sec) {
    /*
      Since both of "tv1->tv_sec" and "tv2->tv_sec" are non-negative and
      integral, "tmp.tv_sec" will be non-negative, integral, and not greater
      than "tv1->tv_sec", thus it can be represented exactly. Both of
      "tv1->tv_usec" and "tv2->tv_usec" are in [0, 999999], the difference will
      be in [-999999, 999999], which can be represented by a "long".
    */
    tmp.tv_sec = tv1->tv_sec - tv2->tv_sec;
    tmp.tv_usec = tv1->tv_usec - tv2->tv_usec;
    ret = 0;

    if (0L > tmp.tv_usec) {
      ret = -1;

      if (tmp.tv_sec > (time_t)0) {
        --tmp.tv_sec;
        tmp.tv_usec += 1000000L;
        ret = 0;
      }
    }
  }

  if (!ret && result)
    *result = tmp;

  return ret;
}
