/* $Id: timer.h,v 1.8 2005/06/04 21:50:12 lacos Exp $ */

#ifndef TIMER_H
#  define TIMER_H
#  include <sys/time.h> /* struct timeval */
#  include <stdlib.h>   /* ldiv_t */

/*-----------------------------------------------------------------------------
  This set of functions manipulates static objects, and is not meant for
  reentrant/multithreaded programs.
*/


/*-----------------------------------------------------------------------------
  Check if "time_t" is an integral type.

  Call this function first.

  Return value:
     0 if it is.
    -1 if it's not. No other functions declared below must be called.
*/
int check_time_t(void);


/*-----------------------------------------------------------------------------
  This function implements the initialization part of the algorithm.

  Parameters:

    current_time: The current time, obtained with "gettimeofday()".

             bps: Transfer rate (the "K" constant, see the implementation
                  notes). This value must be positive and at most "LONG_MAX/2 +
                  1", otherwise the results are undefined.
*/
void timer_init(const struct timeval *current_time, long bps);


/*-----------------------------------------------------------------------------
  Maximum difference allowed between C[I] and C[I-1] (I >= 0), in seconds, as a
  long.

  1800 means half an hour, which should be big enough for any timeout.
  Expressed in microseconds, it can still be represented in a "long". It can
  also be converted to "time_t", since "select()" is guaranteed to support
  timeouts of at least 31 days.
*/

#  define MAXCDIFFL 1800L


/*-----------------------------------------------------------------------------
  This implements the sleeping part of the algorithm's loop.

  Event #I must be executed by the caller before calling this function with
  C[I] and L[I] (I >= 0).

  Parameters:

          c: C[I], "now". It is the caller's responsibility to ensure that the
             time passed since the last call to this function (or to
             "timer_init()", if this function was not yet called) is in [0,
             MAXCDIFFL] seconds. Otherwise (time skew, SIGSTOP), or if the
             previous call was interrupted, a nonzero value must be passed as
             "restart".

          l: L[I], the nominal length of event #I in microseconds. See the
             implementation notes. Must be positive and normalized, and also
             smaller than 31 days (2,678,400,000,000 microseconds). Note that
             the minimum value of "LONG_MAX" mandated by XSHv42 is below this
             value.

    restart: See "c".

  Return value:
     0 if successful.
    -1 if "select()" failed. "Errno" is set by "select()". If you need to call
       this function again, reinitialize the timer using a nonzero value for
       the "restart" parameter, or by calling "timer_init()".
*/
int timer_adv(const struct timeval *c, const ldiv_t *l, int restart);


/*-----------------------------------------------------------------------------
  Calculate "tv1 - tv2" and store the result into "result".

  The caller must ensure that "tv1" and "tv2" are valid and normalized.
  Otherwise, the behaviour is undefined.

  Return value:
     0 if "tv1" >= "tv2". If "result" is not NULL, then the difference is
       assigned to it.
    -1 otherwise. "Result" is unchanged.
*/
int tv_sub(struct timeval *result, const struct timeval *tv1,
    const struct timeval *tv2);

#endif
