/* $Id: timer_test.c,v 1.14 2005/06/04 21:50:12 lacos Exp $ */

#define _XOPEN_SOURCE
#define _XOPEN_SOURCE_EXTENDED 1 /* enable X/Open UNIX Extension */

#include <stdio.h>    /* fprintf() */
#include <signal.h>   /* sigaction() */
#include <string.h>   /* memset() */
#include <errno.h>    /* errno */

#include "timer.h"    /* check_time_t() */
#include "char8bit.h" /* 8 == CHAR_BIT */


static void sigcont_handle(int sig)
{
}


static void signal_setup(void)
{
  struct sigaction sa;

  sa.sa_handler = sigcont_handle;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = 0;
  sigaction(SIGCONT, &sa, 0);

  {
    sigset_t mask;

    sigemptyset(&mask);
    sigaddset(&mask, SIGCONT);
    sigprocmask(SIG_BLOCK, &mask, 0);
  }

  sa.sa_handler = SIG_IGN;
  sigaction(SIGPIPE, &sa, 0);
  sigaction(SIGXFSZ, &sa, 0);
}


static int sigcont_fetch(void)
{
  sigset_t pending;

  sigemptyset(&pending);
  sigpending(&pending);
  if (sigismember(&pending, SIGCONT)) {
    sigset_t mask;

    sigemptyset(&mask);
    sigaddset(&mask, SIGCONT);
    sigprocmask(SIG_UNBLOCK, &mask, 0);
    sigprocmask(SIG_BLOCK, &mask, 0);
    return 1;
  }
  return 0;
}


int main(int argc, char **argv)
{
  int ret = EXIT_FAILURE;

  if (-1 == check_time_t())
    fprintf(stderr, "%s: time_t is not an integral type, unsupported"
        " plaform\n", argv[0]);
  else
    if (3 != argc)
      fprintf(stderr, "usage: %s TIMEOUT BPS\n", argv[0]);
    else {
      long tout = strtol(argv[1], 0, 0);

      /* None of "LONG_MIN", "LONG_MAX", 0 fall into [1, "MAXCDIFFL"]. */
      if (tout > MAXCDIFFL || tout < 1L)
        fprintf(stderr, "%s: invalid TIMEOUT: \"%s\"\n", argv[0], argv[1]);
      else {
        long bps = strtol(argv[2], 0, 0);

        /*
          None of "LONG_MIN", "LONG_MAX", 0 fall into [1, "LONG_MAX/2 + 1"].
        */
        if (bps > (LONG_MAX / 2L + 1L) || bps < 1L)
          fprintf(stderr, "%s: invalid BPS: \"%s\"\n", argv[0], argv[2]);
        else {
          /* "2048L * 1000000L < LONG_MAX", on all UNIX 95 platforms. */
          char buf[2048];
          struct timeval tout_tv,
              prev,
              last_ack;

          /* This is the place where we use 8 == "CHAR_BIT". */
          memset(buf, '1', sizeof buf);

          /* This is safe, "tout" being in [1, "MAXCDIFFL"]. */
          tout_tv.tv_sec = tout;
          tout_tv.tv_usec = 0L;

          /* C[-1] := current_time */
          gettimeofday(&prev, 0);
          last_ack = prev;

          /* "Time_t" is an integer. */
          srand(prev.tv_sec);

          signal_setup();
          ret = EXIT_SUCCESS;
          timer_init(&prev, bps);
          while (EXIT_SUCCESS == ret) {
            int event_bytes = 1024 + rand() % 1025;

            /* L[I] := exec_event(I) */
            fprintf(stdout, "%.*s", event_bytes, buf);
            if (EOF == fflush(stdout)) {
              fprintf(stderr, "%s: fflush(): %s\n", argv[0], strerror(errno));
              ret = EXIT_FAILURE;
            }
            else {
              struct timeval now;
              int recvd,
                  restart = 0;

              /*
                Receive. If at least one ACK was received, note the fact here.
                For now, just pretend.
              */
              recvd = 1;

              /* C[I] := current_time */
              gettimeofday(&now, 0);

              if (sigcont_fetch() || -1 == tv_sub(0, &now, &prev)) {
                if (0 > fprintf(stderr, "%s: restarting timer\n", argv[0]))
                  ret = EXIT_FAILURE;
                else
                  last_ack = now;
                  restart = 1;
              }

              /*
                "Prev" should be set to "now" here, but we save that assignment
                until we know that we will continue the loop. The only place
                where "prev" is read from is the one above. Thus, the general
                formula here is: "now" >= "prev" >= "last_ack".
              */

              if (EXIT_SUCCESS == ret) {
                struct timeval since_last_ack;

                ret = EXIT_FAILURE;

                tv_sub(&since_last_ack, &now, &last_ack);
                if (-1 == tv_sub(0, &tout_tv, &since_last_ack))
                  fprintf(stderr, "%s: timeout\n", argv[0]);
                else {
                  /*
                    The multiplication is safe, see "event_bytes" and "buf".
                  */
                  ldiv_t event_length
                      = ldiv((long)event_bytes * 1000000L, bps);

                  /*
                    SIGCONT: handled with "restart", interruption: handled,
                    negative time skew: handled with restart. Any small
                    positive time skew must be protected with a SIGCONT by the
                    user. Upper limit on CDIFF[I] (I >= 0) is guaranteed,
                    because CDIFF[I] <= "since_last_ack" <= "tout_tv" <=
                    "MAXCDIFFL" seconds. CDIFF[I] is non-negative, see "prev".

                    "Event_length" is positive and normalized. Its maximum
                    value is "sizeof buf * 1000000 / 1" microseconds, which is
                    less than an hour. Of course, it can still lead to timeout.
                  */

                  if (-1 == timer_adv(&now, &event_length, restart))
                    fprintf(stderr, "%s: select(): %s\n", argv[0],
                        strerror(errno));
                  else {
                    if (recvd)
                      last_ack = now;

                    /* I := I + 1 */
                    prev = now;
                    ret = EXIT_SUCCESS;
                  }
                }
              }
            }
          }
        }
      }
    }

  return ret;
}
