/* $Id: udp_copy2.c,v 1.8 2005/06/12 00:24:35 lacos Exp $ */

#define _XOPEN_SOURCE
#define _XOPEN_SOURCE_EXTENDED 1 /* enable X/Open UNIX Extension */

#include "udp_copy2.h"  /* interface; moved to the front for VMS */

#include <arpa/inet.h>  /* htonl() */
#include <signal.h>     /* sig_atomic_t */
#include <errno.h>      /* errno */
#include <stdlib.h>     /* strtol() */
#include <stdio.h>      /* fprintf() */
#include <string.h>     /* strerror() */
#include <sys/socket.h> /* socket() */
#ifdef __VMS
#  include <stropts.h>  /* ioctl() */
#else
#  include <fcntl.h>    /* fcntl() */
#endif
#include <unistd.h>     /* close() */


const char *pname;


u32 *pktno_ntoh(u32 host[PKTNO_U32_NUM], const u32 net[PKTNO_U32_NUM])
{
  u32 *tmp_h = host;
  const u32 *end_h = host + PKTNO_U32_NUM;

  while (tmp_h != end_h)
    *tmp_h++ = ntohl(*net++);

  return host;
}


u32 *pktno_hton(u32 net[PKTNO_U32_NUM], const u32 host[PKTNO_U32_NUM])
{
  u32 *tmp_n = net;
  const u32 *end_n = net + PKTNO_U32_NUM;

  while (tmp_n != end_n)
    *tmp_n++ = htonl(*host++);

  return net;
}


u32 *pktno_incr(u32 inout[PKTNO_U32_NUM])
{
  u32 *ret = 0,
      *tmp = inout;
  const u32 *end = inout + PKTNO_U32_NUM;

  while (!ret && tmp != end)
    if (0U < ++*tmp++) /* Same precedence, right to left associativity. */
      ret = inout;

  return ret;
}


int pktno_sub(u32 result[PKTNO_U32_NUM], const u32 a[PKTNO_U32_NUM],
    const u32 b[PKTNO_U32_NUM])
{
  u32 borrow = 0U;
  const u32 *end_a = a + PKTNO_U32_NUM;

  while (a != end_a) {
    u32 a1 = *a++,
        r1 = a1 - borrow,
        r2 = r1 - *b++;
    borrow = r1 > a1 || r2 > r1;
    if (result)
      *result++ = r2;
  }

  return 0U == borrow ? 0 : -1;
}


static volatile sig_atomic_t sigcont;


static void sigcont_handle(int sig)
{
  sigcont = 1;
}


void signal_init(void)
{
  struct sigaction sa;

  sa.sa_handler = sigcont_handle;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = 0;
  sigaction(SIGCONT, &sa, 0);

  {
    sigset_t mask;

    sigemptyset(&mask);
    sigaddset(&mask, SIGCONT);
    sigprocmask(SIG_BLOCK, &mask, 0);
  }

  sa.sa_handler = SIG_IGN;
  sigaction(SIGPIPE, &sa, 0);
#if !defined (__VMS) || defined(SIGXFSZ)
  /*
    "SIGXFSZ" was not available on "Compaq C V6.2-003 on OpenVMS Alpha V7.3-2".
  */
  sigaction(SIGXFSZ, &sa, 0);
#endif
}


int sigcont_fetch(void)
{
  sigset_t mask;
  int ret;

  sigemptyset(&mask);
  sigaddset(&mask, SIGCONT);
  sigprocmask(SIG_UNBLOCK, &mask, 0);
  sigprocmask(SIG_BLOCK, &mask, 0);
  ret = sigcont;
  sigcont = 0;
  return ret;
}


int sock_init(const struct sockaddr_in *bind_to)
{
  int ret = -1,
      sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

  if (-1 == sock)
    fprintf(stderr, "%s: socket(): %s\n", pname, strerror(errno));
  else {
    if (sock >= (int)FD_SETSIZE) /* FD_SETSIZE can be passed to "select()". */
      fprintf(stderr, "%s: socket descriptor %d too big for an fd_set\n",
          pname, sock);
    else {
#ifdef __VMS
      int opts = 1;

      if (-1 == ioctl(sock, FIONBIO, &opts))
        fprintf(stderr, "%s: ioctl(): %s\n", pname, strerror(errno));
#else
      int opts = fcntl(sock, F_GETFL);

      if (-1 == opts || -1 == fcntl(sock, F_SETFL, opts | O_NONBLOCK))
        fprintf(stderr, "%s: fcntl(): %s\n", pname, strerror(errno));
#endif
      else
        if (!bind_to)
          ret = sock;
        else
          if (-1 == bind(sock, (const struct sockaddr *)bind_to,
              sizeof *bind_to))
            fprintf(stderr, "%s: bind(): %s\n", pname, strerror(errno));
          else
            if (0U != bind_to->sin_port)
              ret = sock;
            else {
              struct sockaddr_in local;
              size_t addr_len = sizeof local;

              if (-1 == getsockname(sock, (struct sockaddr *)&local,
                  &addr_len))
                fprintf(stderr, "%s: getsockname(): %s\n", pname,
                    strerror(errno));
              else
                if (0 < fprintf(stderr, "%s: listening on port %u\n", pname,
                    (unsigned)ntohs(local.sin_port)))
                  ret = sock;
            }
    }

    if (-1 == ret)
      close(sock);
  }

  return ret;
}


int pkt_wait(int sock, const struct timeval *timeout)
{
  fd_set rd;
  sigset_t mask;
  int ret = 1;

  FD_ZERO(&rd);
  sigemptyset(&mask);
  sigaddset(&mask, SIGCONT);

  while (1 == ret) {
    struct timeval tmp = *timeout;
    int sel;

    FD_SET(sock, &rd);
    sigprocmask(SIG_UNBLOCK, &mask, 0);
    sigcont = 0; /* Drop any pending or previous SIGCONT here. */
    sel = select(sock + 1, &rd, 0, 0, &tmp);
    sigprocmask(SIG_BLOCK, &mask, 0);

    switch (sel) {
      case 0:
        if ((sig_atomic_t)0 == sigcont) {
          fprintf(stderr, "%s: timeout\n", pname);
          ret = -1;
        }
        /*
          Otherwise, repeat. On "OSF1 V4.0 1091 alpha", if the process is
          stopped inside "select()", and is continued only after the specified
          timeout has passed (in real time), "select()" returns 0. In this
          case, similar to the one with EINTR, we need to re-"select()".
          However, timeout is treated as success, and "select()" may have
          modified "tmp" and certainly did clear "rd". That's why they are
          reinitialized at the head of the loop. Yes, there is a window right
          before "select()", and another window right after "select()", where a
          SIGCONT can fool this logic, meaning that "select()" will be
          restarted even when a "real" timeout happened. Big deal.
        */
        break;

      case -1:
        if (EINTR != errno) {
          fprintf(stderr, "%s: select(): %s\n", pname, strerror(errno));
          ret = -1;
        }
        /*
          Otherwise, repeat. The safe thing is to reinitialize "tmp" and "rd"
          at the head of the loop.
        */
        break;

      default:
        ret = 0;
        break;
    }

    if (1 == ret && 0 > fprintf(stderr, "%s: restarting select()\n", pname))
      ret = -1;
  }

  return ret;
}


int long_parse(long *result, long min, long max, const char *str)
{
  char *endptr;

  errno = 0;
  *result = strtol(str, &endptr, 0);
  if (errno || *endptr || str == endptr || *result < min || *result > max)
    return -1;

  return 0;
}


int check_transient(int errcode)
{
  switch (errcode) {
    case EWOULDBLOCK:
#if EAGAIN != EWOULDBLOCK
    case EAGAIN:
#endif
    case EHOSTUNREACH:
    case ENETDOWN:
    case ENETUNREACH:
    case ENOBUFS:
    case ENOMEM:
#if !defined(__VMS) || defined(ENOSR)
    /*
      "ENOSR" was not available on "Compaq C V6.2-003 on OpenVMS Alpha V7.3-2".
    */
    case ENOSR:
#endif
      return 0;
  }

  return -1;
}
