/* $Id: udp_copy2.h,v 1.17 2005/06/12 00:24:35 lacos Exp $ */

#ifndef UDP_COPY2_H
#  define UDP_COPY2_H

#  if defined(__VMS) && defined(BROKEN_VMS_TYPES)
/*-----------------------------------------------------------------------------
  On "Compaq C V6.2-003 on OpenVMS Alpha V7.3-2", defining "_XOPEN_SOURCE" and
  "_XOPEN_SOURCE_EXTENDED=1" breaks some system headers (<netinet/in.h> is one
  of them for sure). These "typedef"s below fix them. "BROKEN_VMS_TYPES" must
  be defined by hand at build time (if it's necessary at all).
*/
typedef char  unsigned  u_char;
typedef short unsigned  u_short;
typedef int   unsigned  u_int;
typedef long  unsigned  u_long;
typedef char           *caddr_t;
#  endif

#  include <netinet/in.h> /* in_port_t */
#  include <sys/time.h>   /* struct timeval */

#  include "char8bit.h"   /* 8 == CHAR_BIT */

#  if (INT_MAX >= 0xFFFFFFFFlu || UINT_MAX > 0xFFFFFFFFlu)
/*-----------------------------------------------------------------------------
  This suffices to prohibit integer promotions on "u32" which would result in
  "int", or in an "unsigned" with greater range.
*/
#    error "INT_MAX >= 0xFFFFFFFFlu || UINT_MAX > 0xFFFFFFFFlu, " \
         "unsupported platform"
#  endif

/*-----------------------------------------------------------------------------
  Basename of the program.
*/
extern const char *pname;


/*-----------------------------------------------------------------------------
  Packet format.

  minimum ethernet MTU = 1500 octets (rfc-894)

  -- IP -- (rfc-791)
  u8  version, internet header length  1 octet
  u8  type of service                  1 octet
  u16 total length                     2 octets
  u16 identificatin                    2 octets
  u16 flags, fragment offset           2 octets
  u8  time to live                     1 octet
  u8  protocol                         1 octet
  u16 header checksum                  2 octets
  u32 source address                   4 octets
  u32 destination address              4 octets
                                      ---------
                                      20 octets

  -- UDP -- (rfc-768)
  u16 source port                      2 octets
  u16 destination port                 2 octets
  u16 packet length                    2 octets
  u16 checksum                         2 octets
                                       --------
                                       8 octets

  -- APP --
  u16 sender_slot_num                  2 octets
  u32 pktnum[2]                        8 octets
                                      ---------
                                      10 octets

  If you redefine these, you must range- and type-check everything affected by
  hand.
*/
#  define PKTNO_U32_NUM 2U
#  define OFS_SSN       0U
#  define OFS_PKTNO     (OFS_SSN   + 2U)
#  define OFS_PAYLD     (OFS_PKTNO + 4U * PKTNO_U32_NUM)
#  define IDX_SSN       0U
#  define IDX_PKTNO     1U
#  define IDX_PAYLD     2U

/* Must be bounded by _XOPEN_IOV_MAX, which is at least 16. */
#  define IDX_ALL       3U

/*-----------------------------------------------------------------------------
  Default payload size. Must be positive and not greater than "32767U -
  OFS_PAYLD".

  Also must not be greater than 2147U (the minimum "LONG_MAX" divided by
  "1000000L" and cast to "unsigned"), because this value, mathematically
  multiplied by one million, must fit into a "long" on all platforms. This is
  required for event length calculation in "udp_send2.c".
*/
#  define PAYLD_U8_NUM  (1500U - 28U - OFS_PAYLD)


/*-----------------------------------------------------------------------------
  Elementary types.
*/
typedef char unsigned u8;
typedef in_port_t u16;
typedef in_addr_t u32;



/*-----------------------------------------------------------------------------
  Packet number functions.
*/

/*-----------------------------------------------------------------------------
  Convert "net" to "host" byte order and vice versa.

  "Host" and "net" must be either the same or disjoint. Word order is
  unchanged.

  Return value: the target is returned.
*/
u32 *pktno_ntoh(u32 host[PKTNO_U32_NUM], const u32 net[PKTNO_U32_NUM]);
u32 *pktno_hton(u32 net[PKTNO_U32_NUM], const u32 host[PKTNO_U32_NUM]);


/*-----------------------------------------------------------------------------
  Increment "inout" by one.

  "Inout" must be provided in host byte order and LSW must be placed first.

  Return value:
    "inout" if successful.
       NULL if "inout" overflowed; now it's zero.
*/
u32 *pktno_incr(u32 inout[PKTNO_U32_NUM]);


/*-----------------------------------------------------------------------------
  Calculate "a - b" and store the result in "result" if "result" is not NULL.

  "A" and "b" must be provided in host byte order. LSW must be placed first.
  Any aliasing is allowed. Partial overlaps are not allowed.

  Return value:
     0 if "a >= b".
    -1 otherwise. "Result", if not NULL, contains the lower bits of the result.
*/
int pktno_sub(u32 result[PKTNO_U32_NUM], const u32 a[PKTNO_U32_NUM],
    const u32 b[PKTNO_U32_NUM]);



/*-----------------------------------------------------------------------------
  Signal functions.
*/

/*-----------------------------------------------------------------------------
  Ignore "SIGPIPE" and "SIGXFSZ". Block "SIGCONT", and set up a simple
  indicator handler for it.
*/
void signal_init(void);


/*-----------------------------------------------------------------------------
  Check whether SIGCONT was pending. Clear the indicator.

  Return value:
     0 if "SIGCONT" was not pending.
     1 if "SIGCONT" was pending.
*/
int sigcont_fetch(void);



/*-----------------------------------------------------------------------------
  Socket functions.
*/

/*-----------------------------------------------------------------------------
  Create a socket (1) for UDP, (2) strictly below FD_SETSIZE, (3) non-blocking.
  If "bind_to" is not NULL, bind the socket to it. Then, if the port number is
  zero, print the OS-chosen port to "stderr".

  Diagnostics are written to "stderr".

  Return value:
           -1 if something failed.
    Otherwise the socket descriptor is returned.
*/
int sock_init(const struct sockaddr_in *bind_to);


/*-----------------------------------------------------------------------------
  Wait for a packet. Write diagnostics to "stderr".

  "Sock" must be non-negative and smaller than "FD_SETSIZE". "Timeout" must not
  be greater than 31 days.

  Return value:
     0 if a packet is ready to be read.
    -1 if there was an error or "select()" timed out.
*/
int pkt_wait(int sock, const struct timeval *timeout);



/*-----------------------------------------------------------------------------
  Misc functions.
*/

/*-----------------------------------------------------------------------------
  Parse the string in "str" into a long. Check whether the result is in ["min",
  "max"].

  Return value:
     0 if successful. "Result" contains the parsed and checked value.
    -1 otherwise. "Result" is indeterminate.
*/
int long_parse(long *result, long min, long max, const char *str);


/*-----------------------------------------------------------------------------
  Check "errcode" against transient networking (send/receive) error codes.

  "EINTR" is not considered transient because "SIGCONT" (the only signal we set
  up to interrupt syscalls) is blocked while sending/receiving.

  Return value:
     0 if the error is transient.
    -1 otherwise.
*/
int check_transient(int errcode);

#endif
