/* $Id: udp_recv2.c,v 1.50 2005/06/13 00:50:39 lacos Exp $ */

#define _XOPEN_SOURCE
#define _XOPEN_SOURCE_EXTENDED 1 /* enable X/Open UNIX Extension */

#include "udp_copy2.h"  /* PAYLD_U8_NUM; moved to the front for VMS */

#include <string.h>     /* strrchr () */
#include <stdio.h>      /* fprintf() */
#include <sys/socket.h> /* AF_INET */
#include <unistd.h>     /* getopt() */
#include <errno.h>      /* errno */
#include <netdb.h>      /* gethostbyname() */
#include <arpa/inet.h>  /* htons() */

#include "timer.h"      /* check_time_t() */
#include "version.h"    /* VERSION */


/* Maximum number of slots. */
#define RECV_SLOTS_MAX ((long unsigned)LONG_MAX > (u32)-1 \
    ? (long unsigned)(u32)-1 : (long unsigned)LONG_MAX)

/*
  Default number of slots. Must be positive and not greater than
  "RECV_SLOTS_MAX".
*/
#define RECV_SLOTS_NUM 2048LU


struct recv_cfg {
  /* Longest time allowed to pass before the next packet arrives. */
  struct timeval timeout;

  long unsigned slots_num; /* Number of slots. */
  unsigned payld_u8_num;   /* Number of octets in a full slot. */

  /*
    Receiver sends first packets at this speed [packets per second], NAT. Zero
    if not initiating.
  */
  long initiate;
  struct sockaddr_in addr; /* Address of peer (initiate mode) or listener. */
};


static int usage(void)
{
  fprintf(stderr,
      "%1$s: Receive datagrams and write them reordered to stdout.\n"
      "Copyright (C) 2005 Laszlo Ersek.\n"
      "Build %2$s.\n"
      "\n"
      "Usage:\n"
      "1. %1$s {option}\n"
      "2. %1$s -h\n"
      "\n"
      "All options are optional in non-initiating mode. \"-H\" and \"-p\"\n"
      "are mandatory when initiating. If all options are valid, last\n"
      "occurrences take effect.\n"
      "\n", pname, VERSION);
  fprintf(stderr,
      "  -t timeout   : Maximum number of seconds to wait for the next\n"
      "                 packet. Allowed: [1, %1$ld], default: %1$ld.\n",
      MAXCDIFFL);
  fprintf(stderr,
      "  -n num_slots : Number of receive slots. Allowed: [1, %lu],\n"
      "                 default: %lu, suggested: four times the number of\n"
      "                 send slots.\n", RECV_SLOTS_MAX, RECV_SLOTS_NUM);
  fprintf(stderr,
      "  -s slot_size : Number of octets a packet (or slot) will hold.\n"
      "                 Allowed: [1, %u], default: %u, suggested: PMTU-%u.\n"
      "                 Must not be smaller than the sender's slot size.\n",
      32767U - OFS_PAYLD, PAYLD_U8_NUM, 28U + OFS_PAYLD);
  fprintf(stderr,
      "  -i pps       : Initiate. Packet exchange will be commenced by the\n"
      "                 receiver sending empty packets to the sender. Useful\n"
      "                 if the receiver is behind a network address\n"
      "                 translating firewall. Initiator packets will be sent\n"
      "                 at pps packets per second. Allowed: [1, %ld],\n"
      "                 default: don't initiate.\n", LONG_MAX / 2L + 1L);
  fprintf(stderr,
      "  -H host      : Destination host when initiating. Listen address\n"
      "                 otherwise, default: INADDR_ANY.\n"
      "  -p port      : Destination port when initiating. Listen port\n"
      "                 otherwise, default: use a port chosen by the OS and\n"
      "                 print it to stderr.\n"
      "  -f           : Force dumping even if stdout is a terminal.\n"
      "  -h           : Print this help and exit successfully.\n");

  return ferror(stderr) ? EXIT_FAILURE : EXIT_SUCCESS;
}


/*-----------------------------------------------------------------------------
  Return value:
    EXIT_FAILURE if something failed. "Recv_cfg" is indeterminate.
    EXIT_SUCCESS if user asked for help. "Recv_cfg" is indeterminate.
              -1 otherwise.
*/
static int args_parse(struct recv_cfg *recv_cfg, int argc, char * const *argv)
{
  static const char optstring[] = ":ht:n:s:i:H:p:f";
  int ret = -2,
      force = 0;


  recv_cfg->timeout.tv_sec  = MAXCDIFFL; /* This is safe, see "MAXCDIFFL". */
  recv_cfg->timeout.tv_usec = 0L;
  recv_cfg->slots_num       = RECV_SLOTS_NUM;
  recv_cfg->payld_u8_num    = PAYLD_U8_NUM;
  recv_cfg->initiate        = 0L;

  /*
    Beside clearing "unused" areas, this sets "sin_port" to 0 and "s_addr" to
    INADDR_ANY (0.0.0.0), because these members are unsigned, fixed width
    integers. Byte order is indifferent for these values.
  */
  memset(&recv_cfg->addr, 0, sizeof recv_cfg->addr);
  recv_cfg->addr.sin_family = AF_INET;

  while (-2 == ret)
    switch (getopt(argc, argv, optstring)) {
      case 'h':
        ret = usage();
        break;

      case 't':
        {
          long tmp;

          if (-1 == long_parse(&tmp, 1L, MAXCDIFFL, optarg)) {
            fprintf(stderr, "%s: invalid timeout \"%s\", use \"-h\" for"
                " help\n", pname, optarg);
            ret = EXIT_FAILURE;
          }
          else
            /* This is safe, "tmp" is bounded by "MAXCDIFFL". */
            recv_cfg->timeout.tv_sec = tmp;
        }
        break;

      case 'n':
        {
          long tmp;

          if (-1 == long_parse(&tmp, 1L, RECV_SLOTS_MAX, optarg)) {
            fprintf(stderr, "%s: invalid num_slots \"%s\", use \"-h\" for"
                " help\n", pname, optarg);
            ret = EXIT_FAILURE;
          }
          else
            recv_cfg->slots_num = tmp;
        }
        break;

      case 's':
        {
          long tmp;

          if (-1 == long_parse(&tmp, 1L, (long)(32767U - OFS_PAYLD), optarg)) {
            fprintf(stderr, "%s: invalid slot_size \"%s\", use \"-h\" for"
                " help\n", pname, optarg);
            ret = EXIT_FAILURE;
          }
          else
            recv_cfg->payld_u8_num = tmp;
        }
        break;

      case 'i':
        if (-1 == long_parse(&recv_cfg->initiate, 1L, LONG_MAX / 2L + 1L,
            optarg)) {
          fprintf(stderr, "%s: invalid pps \"%s\", use \"-h\" for help\n",
              pname, optarg);
          ret = EXIT_FAILURE;
        }
        break;

      case 'H':
        {
          struct hostent *he = gethostbyname(optarg);

          /*
           Don't use "h_errno", it is marked to be withdrawn. "AF_INET" can be
           converted to an "int" since it can be passed to "socket()".
          */
          if (!he || (int)AF_INET != he->h_addrtype) {
            fprintf(stderr, "%s: invalid host \"%s\" (%s), use \"-h\" for"
                " help\n", pname, optarg, he ? "invalid address type"
                : "gethostbyname() failed");
            ret = EXIT_FAILURE;
          }
          else
            memcpy(&recv_cfg->addr.sin_addr, he->h_addr_list[0],
                sizeof(struct in_addr));
        }
        break;

      case 'p':
        {
          long tmp;

          if (-1 == long_parse(&tmp, 1L, 65535L, optarg)) {
            fprintf(stderr, "%s: invalid port \"%s\", use \"-h\" for help\n",
                pname, optarg);
            ret = EXIT_FAILURE;
          }
          else
            recv_cfg->addr.sin_port = htons((in_port_t)tmp);
        }
        break;

      case 'f':
        force = 1;
        break;

      case -1:
        if (optind == argc)
          ret = -1;
        else {
          fprintf(stderr, "%s: excess operands on command line, use \"-h\" for"
              " help\n", pname);
          ret = EXIT_FAILURE;
        }
        break;

      case ':':
        fprintf(stderr, "%s: option \"-%c\" requires an argument, use \"-h\""
            " for help\n", pname, optopt);
        ret = EXIT_FAILURE;
        break;

      case '?':
        fprintf(stderr, "%s: unknown option \"-%c\", use \"-h\" for help\n",
            pname, optopt);
        ret = EXIT_FAILURE;
        break;
    }

  if (-1 == ret) {
    if (recv_cfg->initiate > 0L) {
      ret = EXIT_FAILURE;
      if (INADDR_ANY == recv_cfg->addr.sin_addr.s_addr)
        fprintf(stderr,
            "%s: host not set or set to INADDR_ANY, use \"-h\" for help\n",
                pname);
      else
        if (0U == recv_cfg->addr.sin_port)
          fprintf(stderr, "%s: port not set, use \"-h\" for help\n", pname);
        else
          ret = -1;
    }

    if (-1 == ret && !force && 1 == isatty(STDOUT_FILENO)) {
      fprintf(stderr, "%s: not dumping to a terminal unless forced to, use"
          " \"-h\" for help\n", pname);
      ret = EXIT_FAILURE;
    }
  }

  return ret;
}


/*
  The array of slots is allocated at once in the beginning
  ("recv_pktbuf_init()") and freed at once in the end ("recv_pktbuf_uninit()").

  Each payload is allocated on the fly, just before receiving ("receive()"). It
  is freed in one of three places: (a) immediately, if it won't be kept in the
  ring ("receive()"), (b) after being written out ("recv_pktbuf_flush()"), (c)
  in the end ("recv_pktbuf_uninit()"), if it is left in the ring (in case of an
  error, for example).

  On-the-fly allocation is useful because the payload can be linked into place
  with only one "recvmsg()" and without "memcpy()".
*/
struct recv_slot {
  u8 *payld;             /* Slot unused if and only if NULL. */
  unsigned payld_u8_num; /* Number of payload octets if slot is used. */
};


/*
  Speaking with no range checks and types in mind for a second: a valid, new
  packet with a pktno in "[pktno_next, pktno_next + slots_num)" will be placed
  at "slots[(slot_next + pktno - pktno_next) % slots_num]".
*/
struct recv_pktbuf {
  struct recv_slot *slots; /* Fixed size reordering ring buffer. */
  long unsigned slot_next; /* Current ring buffer start offset. */

  /* Number of first not yet written out packet, host byte order, LSW first. */
  u32 pktno_next[PKTNO_U32_NUM];
};


/*-----------------------------------------------------------------------------
  Return value:
     0 if successful.
    -1 otherwise, "pktbuf" is unchanged.
*/
static int recv_pktbuf_init(struct recv_pktbuf *pktbuf,
    long unsigned slots_num)
{
  if ((size_t)-1 / sizeof(struct recv_slot) < slots_num)
    fprintf(stderr, "%s: size_t overflow while allocating slots (num_slots ="
        " %lu)\n", pname, slots_num);
  else {
    struct recv_slot *s = malloc(slots_num * sizeof(struct recv_slot));

    if (!s)
      fprintf(stderr, "%s: malloc(), while allocating slots (num_slots = %lu):"
          " %s\n", pname, slots_num, strerror(errno));
    else {
      const struct recv_slot *slots_end = s + slots_num;

      pktbuf->slots = s;
      pktbuf->slot_next = 0LU;
      memset(pktbuf->pktno_next, 0, sizeof pktbuf->pktno_next);

      while (s != slots_end)
        s++->payld = 0;

      return 0;
    }
  }

  return -1;
}


/*-----------------------------------------------------------------------------
  Send initiator packets at "pps" packets per second to "target", until it
  responds or a timeout or some other error occurs.

  Return value:
     0 if successful.
    -1 otherwise.
*/
static int nat_init(int sock, const struct recv_cfg *recv_cfg)
{
  int ret = 1;

  /*
    For all I >= 0,
    L[I] usec = (packets in event #I) * 1000000 / (K packets/sec).

    Below, "event_length" is positive, normalized, and not longer than 1
    second.
  */
  const ldiv_t event_length = ldiv(1L * 1000000L, recv_cfg->initiate);
  struct timeval prev,
      start;

  /*
    Drop any pending SIGCONT before starting the timer algorithm. If a SIGCONT
    is pending here, it would unnecessarily restart the timer in the first
    loop. The following "gettimeofday()" is already part of the timer
    algorithm, thus "sigcont_fetch()" must precede it. Otherwise, if the
    process was stopped and continued right after the following
    "gettimeofday()", "sigcont_fetch()" would swallow SIGCONT, and the
    algorithm could time out in the first execution of the loop.
  */
  sigcont_fetch();

  /* C[-1] := current_time */
  gettimeofday(&prev, 0);
  start = prev;

  /* "Recv_cfg->initiate" is positive and at most "LONG_MAX / 2L + 1L". */
  timer_init(&prev, recv_cfg->initiate);

  while (1 == ret) {
    /* L[I] := exec_event(I) */
    {
      struct sockaddr_in originator;
      size_t originator_len = sizeof originator;

      /*
        Originally, the second argument to "recvfrom()" and "sendto()" was "0"
        (NULL), but using "Compaq C V6.2-003 on OpenVMS Alpha V7.3-2", the
        system insists on a valid address, even if the size of the buffer is
        zero. No conditional preprocessing is done, since the current soulution
        doesn't hurt on UNIX either.

        "Vms_dummy" is deliberately not defined as "u8", although the types
        "u8" and "char unsigned" are the same (see "char8bit.h"). Here I want
        to stress that "vms_dummy" is of the smallest and most basic data unit
        (not a "char"!) and not that it's an octet.
      */
      char unsigned vms_dummy;

      if (0 == recvfrom(sock, &vms_dummy, 0, MSG_PEEK,
          (struct sockaddr *)&originator, &originator_len))
        /* Message arrived. */
        if (originator.sin_addr.s_addr == recv_cfg->addr.sin_addr.s_addr
            && originator.sin_port == recv_cfg->addr.sin_port)
          /* From sender. */
          ret = 0;
        else
          /* Drop message and continue. */
          (void)recvfrom(sock, &vms_dummy, 0, 0, 0, 0);
      else
        if (-1 == check_transient(errno)) {
          fprintf(stderr, "%s: recvfrom(): %s\n", pname, strerror(errno));
          ret = -1;
        }
    }

    if (1 == ret) {
      char unsigned vms_dummy;

      /* Tolerate temporary errors. */
      if (-1 == sendto(sock, &vms_dummy, 0, 0,
          (const struct sockaddr *)&recv_cfg->addr, sizeof recv_cfg->addr)
          && -1 == check_transient(errno)) {
        fprintf(stderr, "%s: sendto(): %s\n", pname, strerror(errno));
        ret = -1;
      }
      else {
        struct timeval now;
        int restart = 0;

        /* C[I] := current_time */
        gettimeofday(&now, 0);

        /*
          Fetch any pending SIGCONT first, otherwise a time skew would leave it
          pending until the next execution of the loop, leading to an
          unnecessary double restart of the timer.
        */
        if (sigcont_fetch() || -1 == tv_sub(0, &now, &prev)) {
          if (0 < fprintf(stderr, "%s: restarting timer\n", pname)) {
            start = now;
            restart = 1;
          }
          else
            ret = -1;
        }

        if (1 == ret) {
          struct timeval passed;

          /*
            "Passed" is non-negative, either because of "restart", or because
            "now" >= "prev", but "prev" >= "start".
          */
          tv_sub(&passed, &now, &start);

          if (-1 == tv_sub(0, &recv_cfg->timeout, &passed)) {
            fprintf(stderr, "%s: timeout while initiating\n", pname);
            ret = -1;
          }
          else
            /* 0 <= CDIFF[I] <= "now" - "start" <= "MAXCDIFFL". */
            if (-1 == timer_adv(&now, &event_length, restart)) {
              fprintf(stderr, "%s: select(): %s\n", pname, strerror(errno));
              ret = -1;
            }
            else
              /* I := I + 1 */
              prev = now;
        }
      }
    }
  }

  return ret;
}


/*-----------------------------------------------------------------------------
  Send acknowledgement to "peer".

  The elements before "IDX_PAYLD" in "fields" must be filled in with the
  corresponding application headers of the received message that this ACK will
  be sent for.

  Return value:
     0 if successful.
    -1 if failed.
*/
static int pkt_ack(const struct iovec fields[IDX_PAYLD], int sock,
    const struct sockaddr_in *peer)
{
  struct msghdr msg;

  msg.msg_name       = (void *)peer;           /* Not modified by sendmsg(). */
  msg.msg_namelen    = sizeof *peer;
  msg.msg_iov        = (struct iovec *)fields; /* Not modified by sendmsg(). */
  msg.msg_iovlen     = IDX_PAYLD;              /* Why an "int"... */
  msg.msg_control    = 0;
  msg.msg_controllen = 0U;
  /* Ignored: "msg.msg_flags". */

  if (-1 == sendmsg(sock, &msg, 0) && -1 == check_transient(errno)) {
    fprintf(stderr, "%s: sendmsg(): %s\n", pname, strerror(errno));
    return -1;
  }
  return 0;
}


/*-----------------------------------------------------------------------------
  Write out all reordered packets from the buffer, stopping before the first
  missing one.

  Release all packets which are written. Advance "next" pointers. A packet with
  empty payload is considered the terminating packet.

  Return value:
     1 if successful. Continue processing.
     0 if successful. Terminate processing (end of stream written).
    -1 if writing failed, or the "next packet number" wrapped around without an
       EOF. "Recv_pktbuf" may be modified, but it will be left in a consistent
       state.
*/
static int recv_pktbuf_flush(struct recv_pktbuf *recv_pktbuf,
    const struct recv_cfg *recv_cfg)
{
  int ret = 1;
  u8 *payld;

  while (1 == ret
      /* Assignment. */
      && (payld = recv_pktbuf->slots[recv_pktbuf->slot_next].payld)) {
    size_t all = recv_pktbuf->slots[recv_pktbuf->slot_next].payld_u8_num,
        written = 0U;
    ssize_t cur;

    if (0U == all)
      ret = 0;

    /*
      No buffering or synchronous I/O multiplexing here. If the consumer is
      never slower than we are, then the loop below won't block. If the
      consumer is not slower than we are ON THE AVERAGE, an external buffering
      program can be used to inhibit blocking. If the consumer is slower than
      we are, we will block with or without multiplexing or an external buffer.
      (Assuming that the buffer has a fixed maximum size.)
    */
    while (
        /* Two assignments. */
        -1 != (cur = write(STDOUT_FILENO, payld + written, all - written))
        && all != (written += (size_t)cur))
      ;

    /* We expected something blocking. */
    if (-1 == cur) {
      fprintf(stderr, "%s: write(): %s\n", pname, strerror(errno));
      ret = -1;
    }
    else {
      /* Slot written, release it and advance. */
      free(payld);
      recv_pktbuf->slots[recv_pktbuf->slot_next].payld = 0;
      if (++recv_pktbuf->slot_next == recv_cfg->slots_num)
        recv_pktbuf->slot_next = 0LU;

      /* Don't complain about wraparound if this was the EOF packet. */
      if (!pktno_incr(recv_pktbuf->pktno_next) && 1 == ret) {
        fprintf(stderr, "%s: packet number wraparound\n", pname);
        /*
          We have written out all packets up to and including this one. We
          haven't accepted any packet with a higher pktno than the pktno of
          this packet, because this is the highest pktno possible. These two
          facts together mean that now all slots are empty.
          "Recv_pktbuf->pktno_next" just got reset to zero.
          "Recv_pktbuf->slot_next" has advanced to a valid position, as usual.
          All together, this leaves "recv_pktbuf" in a valid state. Not that
          we'd use it for anything after this point (except uninitialization).
        */
        ret = -1;
      }
    }
  }

  return ret;
}


/*-----------------------------------------------------------------------------
  Analyze the packet having full application headers based on named application
  headers and optional payload.

  "Iov_base" members in "fields" must be correctly aligned for casts.
  "Fields[IDX_PAYLD].iov_len" must contain the actual length of the payload.
  "Pkt_drop" must be nonzero on input.

  Return value:
     1 if successful and processing must continue. "Pkt_drop" is set to a
       nonzero value if the payload should be freed. It is set to 0 if the
       payload shouldn't be freed.
     0 if successful and the whole stream is written. In this case, all packets
       are freed, thus "pkt_drop" will be set to zero.
    -1 if something failed and processing should abort. Output parameters may
       be changed, but they are left consistent and meaningful FOR CLEANUP.
*/
static int pkt_zoom(int *pkt_drop, struct recv_pktbuf *recv_pktbuf,
    const struct recv_cfg *recv_cfg, const struct iovec fields[IDX_ALL],
    int sock, const struct sockaddr_in *peer)
{
  /* Default settings: continue processing, ACK packet, drop packet. */
  int ret = 1,
    needs_ack = 1;

  /* Offset relative to "recv_pktbuf->pktno_next". */
  u32 pktno_ofs[PKTNO_U32_NUM];

  /*
    The sender must ensure that
    - if pktno1 < pktno2, then payload1 precedes payload2,
    - if pktno1 = pktno2, then payload1 equals payload2.
    Otherwise (for example: pktno wraps around) there is no guarantee that the
    output stream won't be garbled.
  */
  if (-1 != pktno_sub(pktno_ofs, pktno_ntoh(pktno_ofs, /* ugly reuse */
      fields[IDX_PKTNO].iov_base), recv_pktbuf->pktno_next)) {
    /* Packet not yet written out. */

    u32 pktno_ofs_max[PKTNO_U32_NUM] = { 0U };

    /*
      This is safe, see "RECV_SLOTS_MAX", "RECV_SLOTS_NUM" and "args_parse()".
    */
    pktno_ofs_max[0] = recv_cfg->slots_num;

    if (-1 == pktno_sub(0, pktno_ofs, pktno_ofs_max)) {
      /*
        Packet in range. Slot offset is in "pktno_ofs[0]", see "pktno_ofs_max".
      */
      long unsigned slot_ofs = pktno_ofs[0],
          /*
            Both "recv_pktbuf->slot_next" and "slot_ofs" are smaller than
            "recv_cfg->slots_num". If their sum is smaller than the latter
            (which is true, for example, if "1LU == recv_cfg->slots_num" and we
            reach this point), add them. In this case, there will be no "long
            unsigned" overflow either. (All participants are of type "long
            unsigned".)

            Otherwise, the mathematical sum falls into "[recv_cfg->slots_num, 2
            * recv_cfg->slots_num - 2]" (not an empty set, here even
            "recv_cfg->slots_num >= 2LU" holds). The ring buffer position wraps
            around into [0, "recv_cfg->slots_num - 2LU"] by subtracting
            "recv_cfg->slots_num". This latter set, which contains the result,
            is a subset of "long unsigned"'s domain, thus we don't need the
            extra bit, which could be required for representing the
            intermediate mathematical sum, even while adding and subtracting.

            The important part is to decide about the subtraction with a
            comparison before any truncation modulo "ULONG_MAX + 1" happens.
          */
          slot_abs = (recv_cfg->slots_num - recv_pktbuf->slot_next > slot_ofs)
          ? (recv_pktbuf->slot_next + slot_ofs)
          : (recv_pktbuf->slot_next + slot_ofs - recv_cfg->slots_num);

      if (!recv_pktbuf->slots[slot_abs].payld) {
        /* Paket not yet received, don't drop. */
        *pkt_drop = 0;
        recv_pktbuf->slots[slot_abs].payld        = fields[IDX_PAYLD].iov_base;
        recv_pktbuf->slots[slot_abs].payld_u8_num = fields[IDX_PAYLD].iov_len;
      }
    }
    else
      /* Packet behind number of slots. We'll need it later, no ACK. */
      needs_ack = 0;
  }

  if (needs_ack) {
    if (-1 == pkt_ack(fields, sock, peer))
      ret = -1;
    else
      if (!*pkt_drop)
        ret = recv_pktbuf_flush(recv_pktbuf, recv_cfg);
  }

  return ret;
}


/*-----------------------------------------------------------------------------
  Return value:
     0 if successful (whole stream written out).
    -1 otherwise.
*/
static int receive(struct recv_pktbuf *recv_pktbuf, int sock,
    const struct recv_cfg *recv_cfg)
{
  struct sockaddr_in peer;
  int ret = 1;

  if (recv_cfg->initiate)
    peer = recv_cfg->addr; /* Not INADDR_ANY, see "args_parse()". */
  else
    peer.sin_addr.s_addr = INADDR_ANY;

  /*
    Kind of cheating. Timeout is only measured while "actively" waiting for a
    packet (within "select()"). This allows some drift (processor time spent
    outside of "select()" is not accounted for, which leaves a bit more time
    for the next packet to arrvie), but for reception, it is not critical. Also
    note that inbound packets of any kind can keep the process from timing out.
  */
  while (1 == ret)
    /*
      It is guaranteed by "listener_init()" that "sock" is non-negative and
      smaller than "FD_SETSIZE". It is guaranteed by "args_parse()" that
      "recv_cfg->timeout" is bounded by "MAXCDIFFL", which is less than 31
      days.
    */
    if (-1 == pkt_wait(sock, &recv_cfg->timeout))
      ret = -1;
    else {
      /* "Recv_cfg->payld_u8_num" is in [1U, 32767U). See "SSIZE_MAX". */
      u8 *payld = malloc(recv_cfg->payld_u8_num);

      if (!payld) {
        fprintf(stderr, "%s: malloc(): %s\n", pname, strerror(errno));
        ret = -1;
      }
      else {
        struct iovec sct[IDX_ALL];
        u16 ssn_net;
        u32 pktno_net[PKTNO_U32_NUM];
        struct msghdr msg;
        struct sockaddr_in originator;
        ssize_t recvd;
        int pkt_drop = 1;

        /*
          It is ensured by "args_parse()" that "recv_cfg->payld_u8_num" plus
          all application headers together occupy no more than 32767U bytes,
          which is not greater than SSIZE_MAX.
        */
        sct[IDX_SSN  ].iov_base = &ssn_net;
        sct[IDX_SSN  ].iov_len  = sizeof ssn_net;
        sct[IDX_PKTNO].iov_base = pktno_net;
        sct[IDX_PKTNO].iov_len  = sizeof pktno_net;
        sct[IDX_PAYLD].iov_base = payld;
        sct[IDX_PAYLD].iov_len  = recv_cfg->payld_u8_num;

        msg.msg_name       = &originator;
        msg.msg_namelen    = sizeof originator;
        msg.msg_iov        = sct;
        msg.msg_iovlen     = sizeof sct / sizeof sct[0]; /* Why an "int"... */
        msg.msg_control    = 0;
        msg.msg_controllen = 0U;
        /*
          Ignored on input: "msg.msg_flags", but zero it out, because it
          doesn't happen automatically with "Compaq C V6.2-003 on OpenVMS Alpha
          V7.3-2".
        */
        msg.msg_flags      = 0;

        /*
          "EWOULDBLOCK", "EAGAIN": should not happen because of "pkt_wait()".
        */
        recvd = recvmsg(sock, &msg, 0);

        if (-1 == recvd) {
          /*
            Since the time spent here is not accounted for, we can't tolerate
            any errors here, otherwise an endless loop would be possible. This
            strictness is not worse than the one against "malloc()"'s failure,
            which, on the other hand, seems natural. The situation when
            "select()" doesn't sleep at all in "pkt_wait()" because of a
            pending error for "sock", is also handled by this check.
          */
          fprintf(stderr, "%s: recvmsg(): %s\n", pname, strerror(errno));
          ret = -1;
        }
        else
          /*
            Don't care about packets
            too big:             msg.msg_flags & MSG_TRUNC
            with ancillary data: msg.msg_flags & MSG_CTRUNC
            with OOB data (*):   msg.msg_flags & MSG_OOB
            with EOR info (**):  msg.msg_flags & MSG_EOR
            too small:           recvd < OFS_PAYLD
            *  should not be possible, was not requested
            ** should not be possible, IPPROTO_UDP
          */
          if (0 == (msg.msg_flags
              & (MSG_TRUNC | MSG_CTRUNC | MSG_OOB | MSG_EOR))
              && (size_t)recvd >= OFS_PAYLD) {
            /* Headers are complete. Check source. */
            if (INADDR_ANY == peer.sin_addr.s_addr)
              peer = originator; /* Not initiating, first packet seen. */
            if (peer.sin_addr.s_addr == originator.sin_addr.s_addr
                && peer.sin_port == originator.sin_port) {
              sct[IDX_PAYLD].iov_len = (size_t)recvd - OFS_PAYLD;
              ret = pkt_zoom(&pkt_drop, recv_pktbuf, recv_cfg, sct, sock,
                  &peer);
            }
          }

        if (pkt_drop)
          free(payld);
      }
    }

  return ret;
}


/*-----------------------------------------------------------------------------
  Do not use "pktbuf" after calling this.
*/
static void recv_pktbuf_uninit(struct recv_pktbuf *pktbuf,
    long unsigned slots_num)
{
  struct recv_slot *s = pktbuf->slots;
  const struct recv_slot *slots_end = s + slots_num;

  while (s != slots_end)
    free(s++->payld);

  free(pktbuf->slots);
}


int main(int argc, char **argv)
{
  int ret = EXIT_FAILURE;

  pname = strrchr(argv[0], '/');
  pname = pname ? pname + 1 : argv[0];
  signal_init();

#ifndef __VMS
  /*
    Close "stdin", ignore return value. For explanation of the conditional
    inclusion see the corresponding area in "udp_send2.c".
  */
  (void)fclose(stdin);
#endif

  if (-1 == check_time_t())
    fprintf(stderr, "%s: time_t is not an integral type, unsupported"
        " platform\n", pname);
  else {
    struct recv_cfg recv_cfg;

    ret = args_parse(&recv_cfg, argc, argv);
    if (-1 == ret) {
      struct recv_pktbuf recv_pktbuf;

      ret = EXIT_FAILURE;
      if (-1 != recv_pktbuf_init(&recv_pktbuf, recv_cfg.slots_num)) {
        int sock = sock_init(recv_cfg.initiate ? 0 : &recv_cfg.addr);

        if (-1 != sock) {
          if ((!recv_cfg.initiate || -1 != nat_init(sock, &recv_cfg))
              && -1 != receive(&recv_pktbuf, sock, &recv_cfg)) {
#ifndef __VMS
            /*
              See "fclose()" above. On UNIX, it's advisable to "close()"
              "STDOUT_FILENO" in a controlled way, so why not do it like this?
            */
            if (EOF == fclose(stdout))
              fprintf(stderr, "%s: fclose(): %s\n", pname, strerror(errno));
            else
#endif
              ret = EXIT_SUCCESS;
          }
          close(sock);
        }
        recv_pktbuf_uninit(&recv_pktbuf, recv_cfg.slots_num);
      }
    }
  }

  return ret;
}
