/* $Id: udp_send2.c,v 1.27 2005/06/13 00:50:39 lacos Exp $ */

#define _XOPEN_SOURCE
#define _XOPEN_SOURCE_EXTENDED 1 /* enable X/Open UNIX Extension */

#include "udp_copy2.h"  /* PKTNO_U32_NUM; moved to the front for VMS */

#include <string.h>     /* strerror() */
#include <stdio.h>      /* fprintf() */
#include <sys/socket.h> /* AF_INET */
#include <unistd.h>     /* STDIN_FILENO */
#include <errno.h>      /* errno */
#include <netdb.h>      /* gethostbyname() */
#include <arpa/inet.h>  /* htons() */

#include "timer.h"      /* check_time_t() */
#include "version.h"    /* VERSION */


/* Maximum number of slots. */
#define SEND_SLOTS_MAX 65536LU

/*
  Default number of slots. Must be positive and not greater than
  "SEND_SLOTS_MAX".
*/
#define SEND_SLOTS_NUM 512LU

/*
  Maximum payload size. Two requirements: (1) together with headers, packet
  must not be greater than the minimum "SSIZE_MAX", that is, "(ssize_t)32767";
  (2) payload size, mathematically multiplied by one million, must fit into a
  "long". Reason for (1): "sendmsg()". Reason for (2): event length
  calculation. (1) also guarantees that payload size can be represented as an
  "unsigned".
*/
#define SEND_PAYLD_U8_MAX ((long unsigned)(LONG_MAX / 1000000L) \
    < (long unsigned)(32767U - OFS_PAYLD) ? (unsigned)(LONG_MAX / 1000000L) \
    : (32767U - OFS_PAYLD))


struct send_cfg {
  /* Longest time allowed to pass before the next ack arrives. */
  struct timeval timeout;

  long unsigned slots_num; /* Number of slots. */
  unsigned payld_u8_num;   /* Number of octets in a full slot. */

  /*
    Nominal length (in time) of sending one full sized packet: L[I] (for all I
    >= 0, because we pretend to send only full sized packets). Expressed in
    microseconds (see "timer_design.txt"): "event_length.quot" determines whole
    microseconds, "event_length.rem" determines "1/bps" microseconds.
  */
  ldiv_t event_length;

  long bps;      /* Sender sends packets at this rate [bytes per second]. */
  int wait_mode; /* Sender waits for receiver to initiate. */

  /* Address of peer if not in wait mode. Address of listener otherwise. */
  struct sockaddr_in addr;
};


static int usage(void)
{
  fprintf(stderr,
      "%1$s: Packetize data read from stdin, send datagrams.\n"
      "Copyright (C) 2005 Laszlo Ersek.\n"
      "Build %2$s.\n"
      "\n"
      "Usage:\n"
      "1. %1$s {option}\n"
      "2. %1$s -h\n"
      "\n"
      "Option \"-b\" is mandatory. Options \"-H\" and \"-p\" are optional if\n"
      "in wait mode, mandatory otherwise. All other options are optional. If\n"
      "all options are valid, last occurrences take effect.\n"
      "\n", pname, VERSION);
  fprintf(stderr,
      "  -b bps       : Send data at bps bytes per second.\n"
      "                 Allowed: [1, %ld].\n", LONG_MAX / 2L + 1L);
  fprintf(stderr,
      "  -t timeout   : Maximum number of seconds allowed to pass before\n"
      "                 the next acknowledgement arrives.\n"
      "                 Allowed: [1, %1$ld], default: %1$ld.\n", MAXCDIFFL);
  fprintf(stderr,
      "  -n num_slots : Number of send slots. Allowed: [1, %lu],\n"
      "                 default: %lu, suggested: one fourth of the number of\n"
      "                 receive slots.\n", SEND_SLOTS_MAX, SEND_SLOTS_NUM);
  fprintf(stderr,
      "  -s slot_size : Number of octets a packet (or slot) will hold.\n"
      "                 Allowed: [1, %u], default: %u, suggested: PMTU-%u.\n"
      "                 Must not be greater than the receiver's slot size.\n",
      SEND_PAYLD_U8_MAX, PAYLD_U8_NUM, 28U + OFS_PAYLD);
  fprintf(stderr,
      "  -w           : Wait mode. Wait for the receiver to initiate the\n"
      "                 transfer with empty packets. Useful if the receiver\n"
      "                 is behind a network address translating firewall.\n"
      "                 Default: normal mode, meaning that the sender starts\n"
      "                 the transfer.\n"
      "  -H host      : In wait mode: listen address, defaults to\n"
      "                 INADDR_ANY. In normal mode: destination host.\n");
  fprintf(stderr,
      "  -p port      : In wait mode: listen port, defaults to an OS-chosen\n"
      "                 port, which is printed to stderr at startup. In\n"
      "                 normal mode: destination port.\n"
      "  -f           : Force packetizing even if stdin is a terminal.\n"
      "  -h           : Print this help and exit successfully.\n");

  return ferror(stderr) ? EXIT_FAILURE : EXIT_SUCCESS;
}


/*-----------------------------------------------------------------------------
  Return value:
    EXIT_FAILURE if something failed. "Send_cfg" is indeterminate.
    EXIT_SUCCESS if user asked for help. "Send_cfg" is indeterminate.
              -1 otherwise.
*/
static int args_parse(struct send_cfg *send_cfg, int argc, char * const *argv)
{
  static const char optstring[] = ":b:t:n:s:wH:p:fh";

  int ret = -2,
      force = 0;

  send_cfg->timeout.tv_sec  = MAXCDIFFL; /* This is safe, see "MAXCDIFFL". */
  send_cfg->timeout.tv_usec = 0L;
  send_cfg->slots_num       = SEND_SLOTS_NUM;
  send_cfg->payld_u8_num    = PAYLD_U8_NUM;
  send_cfg->bps             = 0L;
  send_cfg->wait_mode       = 0;

  /*
    Beside clearing "unused" areas, this sets "sin_port" to 0 and "s_addr" to
    INADDR_ANY (0.0.0.0), because these members are unsigned, fixed width
    integers. Byte order is indifferent for these values.
  */
  memset(&send_cfg->addr, 0, sizeof send_cfg->addr);
  send_cfg->addr.sin_family = AF_INET;

  while (-2 == ret)
    switch (getopt(argc, argv, optstring)) {
      case 'b':
        if (-1 == long_parse(&send_cfg->bps, 1L, LONG_MAX / 2L + 1L,
            optarg)) {
          fprintf(stderr, "%s: invalid bps \"%s\", use \"-h\" for help\n",
              pname, optarg);
          ret = EXIT_FAILURE;
        }
        break;

      case 't':
        {
          long tmp;

          if (-1 == long_parse(&tmp, 1L, MAXCDIFFL, optarg)) {
            fprintf(stderr, "%s: invalid timeout \"%s\", use \"-h\" for"
                " help\n", pname, optarg);
            ret = EXIT_FAILURE;
          }
          else
            /* This is safe, "tmp" is bounded by "MAXCDIFFL". */
            send_cfg->timeout.tv_sec = tmp;
        }
        break;

      case 'n':
        {
          long tmp;

          if (-1 == long_parse(&tmp, 1L, (long)SEND_SLOTS_MAX, optarg)) {
            fprintf(stderr, "%s: invalid num_slots \"%s\", use \"-h\" for"
                " help\n", pname, optarg);
            ret = EXIT_FAILURE;
          }
          else
            send_cfg->slots_num = tmp;
        }
        break;

      case 's':
        {
          long tmp;

          if (-1 == long_parse(&tmp, 1L, (long)SEND_PAYLD_U8_MAX, optarg)) {
            fprintf(stderr, "%s: invalid slot_size \"%s\", use \"-h\" for"
                " help\n", pname, optarg);
            ret = EXIT_FAILURE;
          }
          else
            /* This is safe, see definition of "SEND_PAYLD_U8_MAX". */
            send_cfg->payld_u8_num = tmp;
        }
        break;

      case 'w':
        send_cfg->wait_mode = 1;
        break;

      case 'H':
        {
          struct hostent *he = gethostbyname(optarg);

          /*
           Don't use "h_errno", it is marked to be withdrawn. "AF_INET" can be
           converted to an "int" since it can be passed to "socket()".
          */
          if (!he || (int)AF_INET != he->h_addrtype) {
            fprintf(stderr, "%s: invalid host \"%s\" (%s), use \"-h\" for"
                " help\n", pname, optarg, he ? "invalid address type"
                : "gethostbyname() failed");
            ret = EXIT_FAILURE;
          }
          else
            memcpy(&send_cfg->addr.sin_addr, he->h_addr_list[0],
                sizeof(struct in_addr));
        }
        break;

      case 'p':
        {
          long tmp;

          if (-1 == long_parse(&tmp, 1L, 65535L, optarg)) {
            fprintf(stderr, "%s: invalid port \"%s\", use \"-h\" for help\n",
                pname, optarg);
            ret = EXIT_FAILURE;
          }
          else
            send_cfg->addr.sin_port = htons((in_port_t)tmp);
        }
        break;

      case 'f':
        force = 1;
        break;

      case 'h':
        ret = usage();
        break;

      case -1:
        if (optind == argc)
          ret = -1;
        else {
          fprintf(stderr, "%s: excess operands on command line, use \"-h\" for"
              " help\n", pname);
          ret = EXIT_FAILURE;
        }
        break;

      case ':':
        fprintf(stderr, "%s: option \"-%c\" requires an argument, use \"-h\""
            " for help\n", pname, optopt);
        ret = EXIT_FAILURE;
        break;

      case '?':
        fprintf(stderr, "%s: unknown option \"-%c\", use \"-h\" for help\n",
            pname, optopt);
        ret = EXIT_FAILURE;
        break;
    }

  if (-1 == ret) {
    if (0L == send_cfg->bps) {
      fprintf(stderr, "%s: bps not set, use \"-h\" for help\n", pname);
      ret = EXIT_FAILURE;
    }
    else {
      /*
        This is safe. First, "send_cfg->bps" is in "[1L, LONG_MAX / 2L + 1L]".
        Second, "send_cfg->payld_u8_num" is in "[1L, (long)SEND_PAYLD_U8_MAX]",
        set either by option "-s" or by default. See definitions of
        "SEND_PAYLD_U8_MAX" and "PAYLD_U8_NUM".

        The maximum event length is below 10 hours: "send_cfg->payld_u8_num" is
        smaller than "32767U", "send_cfg->bps" is not smaller than "1L". This
        enables us to put "send_cfg->event_length" into "timer_adv()" later on.
      */
      send_cfg->event_length = ldiv((long)send_cfg->payld_u8_num * 1000000L,
          send_cfg->bps);

      if (!send_cfg->wait_mode) {
        if (INADDR_ANY == send_cfg->addr.sin_addr.s_addr) {
          fprintf(stderr,
              "%s: host not set or set to INADDR_ANY, use \"-h\" for help\n",
                  pname);
          ret = EXIT_FAILURE;
        }
        else
          if (0U == send_cfg->addr.sin_port) {
            fprintf(stderr, "%s: port not set, use \"-h\" for help\n", pname);
            ret = EXIT_FAILURE;
          }
      }

      if (-1 == ret && !force && 1 == isatty(STDIN_FILENO)) {
        fprintf(stderr, "%s: not reading from a terminal unless forced to,"
            " use \"-h\" for help\n", pname);
        ret = EXIT_FAILURE;
      }
    }
  }

  return ret;
}


/*
  All slot descriptors ("send_pktbuf.slots") and slot data
  ("send_pktbuf.slotdata") are allocated at once in the beginning
  ("send_pktbuf_init()"). Still in the same step, the slot data buffer is
  prefilled from "stdin". Used slots are organized (via "send_slot.next",
  "send_slot.prev") into a doubly linked list (which is, actually, an array,
  for random access). If, because of an early EOF, some slots remain empty,
  they are not linked to the list. All slots and slot data are released in one
  step in the end ("send_pktbuf_uninit()").

  Slot data are allocated in one common, continuous chunk. Destination interval
  for data to be read is always known beforehand.

  The main loop ("snd()" calling "pkt_pump()") iterates over the slots in the
  list. After checking for any pending, random acknowledgements
  ("pkt_ckack()"), it sends the payload and administrative stuff found in the
  current slot (designated by "send_pktbuf.slot_next") to the receiver. Any
  slot can be resent several times without modification if no ack arrives for
  it before the loop hits it again and again.

  Each ack carries back the packet number and (for quick lookup) the sender
  slot number of its request. If an ack is valid and fresh, the payload in the
  corresponding slot is replaced with new data from "stdin", its length is
  registered into "send_slot.payld_u8_num", the packet number
  ("send_slot.pktno_net") is replaced with the next packet number in the whole
  stream ("send_pktbuf.pktno_next"); then, the iteration continues. Invalid or
  old acks are dropped.

  If there is nothing more to read from "stdin", slots are unlinked as acked.
  Instead of unlinking the last slot, it is filled with an EOF packet (no new
  payload, "send_slot.payld_u8_num" is set to zero). Upon receiving an ack for
  the EOF packet, the loop terminates. Note that the receiver will only send an
  ack for the first EOF packet it receives, thus, if that ack gets lost, the
  sender will time out. This is no problem - in any case, the success of the
  transfer must be determined, by some method, on the receiver's side.

  "Send_slot.ssn_net" is constant after initialization, it is precomputed for
  simplicity. "Send_slot.payld" is also precomputed, but it is set to NULL when
  the slot is unlinked.

  Note that the timer is able to handle different event lengths, we just don't
  use that capability and pretend to send full packets without exception. (The
  only exceptions would be: the EOF packet, and the (possibly not full) last
  data packet.)

  When the standard input is a regular file, new data can appear on it even
  after an EOF condition. Basically, this is no problem, the algorithm runs on,
  just with fewer slots, e.g. worse performance.
*/
struct send_slot {
  /*
    Fields are meaningful only if the slot is used, i.e. if it can be reached
    by traversing the list, starting with "send_pktbuf.slot_next". This is also
    equivalent to "payld" being non-NULL.
  */
  u8 *payld;              /* Payload. Points into "slotdata" below if used. */
  unsigned payld_u8_num;  /* Number of payload octets in this slot. */
  struct send_slot *next, /* Next slot. */
      *prev;              /* Previous slot. */
  u16 ssn_net;            /* Sender slot number, network byte order. */

  /* Pktno of the packet in this slot, network byte order, LSW first. */
  u32 pktno_net[PKTNO_U32_NUM];
};


struct send_pktbuf {
  struct send_slot *slots, /* Array of slots. */
      *slot_next;          /* Next slot to send. */
  u8 *slotdata;            /* Data for all slots. */

  /* Next pktno to assign, host byte order, LSW first. */
  u32 pktno_next[PKTNO_U32_NUM];
};


/*-----------------------------------------------------------------------------
  Return value:
     0 if successful.
    -1 otherwise. "Send_pktbuf" is unchanged.
*/
static int send_pktbuf_init(struct send_pktbuf *send_pktbuf,
    const struct send_cfg *send_cfg)
{
  /*
    Greater denominator means stricter check. "Send_cfg->payld_u8_num" is in
    "[1U, 32767U)".
  */
  if ((size_t)-1 / (send_cfg->payld_u8_num > sizeof(struct send_slot)
      ? (size_t)send_cfg->payld_u8_num : sizeof(struct send_slot))
      < send_cfg->slots_num)
    fprintf(stderr, "%s: size_t overflow while allocating slots (num_slots ="
        " %lu, slot_size = %u)\n", pname, send_cfg->slots_num,
        send_cfg->payld_u8_num);
  else {
    /* "Send_cfg->slots_num" is in [1LU, 65536LU]. */
    struct send_slot *s = malloc(sizeof(struct send_slot)
        * send_cfg->slots_num);

    if (!s)
      fprintf(stderr, "%s: malloc(), while allocating slots (num_slots = %lu):"
          " %s\n", pname, send_cfg->slots_num, strerror(errno));
    else {
      size_t left = (size_t)send_cfg->payld_u8_num * send_cfg->slots_num;
      u8 *sdata = malloc(left);

      if (!sdata)
        fprintf(stderr, "%s: malloc(), while allocating slot data (num_slots ="
            " %lu, slot_size = %u): %s\n", pname, send_cfg->slots_num,
            send_cfg->payld_u8_num, strerror(errno));
      else {
        size_t rd = 0U;
        ssize_t cur;

        /*
          See the comment on blocking in "udp_recv2.c" at
          "recv_pktbuf_flush()".
        */
        while (0U < left && 0 < (cur = read(STDIN_FILENO, sdata + rd,
            (size_t)SSIZE_MAX < left ? (size_t)SSIZE_MAX : left))) {
          left -= (size_t)cur;
          rd += (size_t)cur;
        }

        if (-1 == cur)
          fprintf(stderr, "%s: read(): %s\n", pname, strerror(errno));
        else {
          const struct send_slot *slot_end = s + send_cfg->slots_num;
          struct send_slot *slot = s,
              *slot_prev = s + (0U == rd ? (size_t)0
                  /*
                    Point "slot_prev" at the last non-empty slot. "Rd <=
                    (size_t)send_cfg->payld_u8_num * send_cfg->slots_num".
                  */
                  : rd / (size_t)send_cfg->payld_u8_num
                  - (size_t)(0U == rd % send_cfg->payld_u8_num));
          u8 *payld = sdata;
          u16 ssn_host = 0U;
          u32 pktno_host[PKTNO_U32_NUM] = { 0U };

          left = rd;
          rd = 0U;

          /*
            Double-link and initialize slots. At least the first slot must be
            initialized, even if no octets could be read from "stdin".
          */
          do {
            unsigned payld_u8_num = left < send_cfg->payld_u8_num
                ? (unsigned)left : send_cfg->payld_u8_num;

            slot->payld = payld;
            slot->payld_u8_num = payld_u8_num;
            slot_prev->next = slot;
            slot->prev = slot_prev;
            slot->ssn_net = htons(ssn_host);
            pktno_hton(slot->pktno_net, pktno_host);

            left -= payld_u8_num;
            slot_prev = slot;
            ++slot;
            payld += payld_u8_num;

            /*
              The loop is not executed more than "send_cfg->slots_num" times,
              that is, the increments below are executed at most 65536 times.
              "Ssn_host" may be reset to zero. "Pktno_host" can't overflow.
            */
            ssn_host = (unsigned)ssn_host + 1U; /* Don't promote to "int". */
            pktno_incr(pktno_host);
          } while (0U < left);

          /* Mark unused slots as such. */
          while (slot != slot_end)
            slot++->payld = 0;

          send_pktbuf->slots = s;
          send_pktbuf->slot_next = s;
          send_pktbuf->slotdata = sdata;
          memcpy(send_pktbuf->pktno_next, pktno_host, sizeof pktno_host);
          return 0;
        }

        free(sdata);
      }

      free(s);
    }
  }

  return -1;
}


/*-----------------------------------------------------------------------------
  Do not use "send_pktbuf" after calling this.
*/
static void send_pktbuf_uninit(struct send_pktbuf *send_pktbuf)
{
  free(send_pktbuf->slots);
  free(send_pktbuf->slotdata);
}


/*-----------------------------------------------------------------------------
  "Ack" is set to a nonzero value if a valid (not necessarily fresh) ACK was
  received. Otherwise, "ack" is not changed.

  Return value:
     1 if successful. Continue processing.
     0 if successful. Terminate processing, EOF packet was ACKed.
    -1 if failed. "Send_pktbuf" must be uninitialized.
*/
static int pkt_ckack(int *ack, struct send_pktbuf *send_pktbuf,
    const struct sockaddr_in *peer, int sock, const struct send_cfg *send_cfg)
{
  int ret = 2;

  while (2 == ret) {
    u16 ssn_net;
    u32 pktno_net[PKTNO_U32_NUM];
    struct sockaddr_in originator;
    struct iovec fields[IDX_PAYLD];
    struct msghdr msg;
    ssize_t recvd;

    fields[IDX_SSN  ].iov_base = &ssn_net;
    fields[IDX_SSN  ].iov_len  = sizeof ssn_net;
    fields[IDX_PKTNO].iov_base = pktno_net;
    fields[IDX_PKTNO].iov_len  = sizeof pktno_net;
    /*
      Sum of all "iov_len" fields is "OFS_PAYLD", smaller than "32767U". See
      first restriction on "PAYLD_U8_NUM".
    */

    msg.msg_name       = &originator;
    msg.msg_namelen    = sizeof originator;
    msg.msg_iov        = fields;
    msg.msg_iovlen     = sizeof fields / sizeof fields[0]; /* "Int"? */
    msg.msg_control    = 0;
    msg.msg_controllen = 0U;
    /*
      Ignored on input: "msg.msg_flags", but zero it out, because it doesn't
      happen automatically with "Compaq C V6.2-003 on OpenVMS Alpha V7.3-2".
    */
    msg.msg_flags      = 0;

    recvd = recvmsg(sock, &msg, 0);

    if (-1 == recvd)
      if (-1 == check_transient(errno)) {
        fprintf(stderr, "%s: recvmsg(): %s\n", pname, strerror(errno));
        ret = -1;
      }
      else
        ret = 1;
    else
      if (originator.sin_addr.s_addr == peer->sin_addr.s_addr
          && originator.sin_port == peer->sin_port
          && OFS_PAYLD == (size_t)recvd && 0 == (msg.msg_flags
          & (MSG_TRUNC | MSG_CTRUNC | MSG_OOB | MSG_EOR))) {
        unsigned ssn = ntohs(ssn_net);

        if (ssn < send_cfg->slots_num) {
          struct send_slot *slot = send_pktbuf->slots + ssn;
          u8 *payld = slot->payld;

          *ack = 1; /* ACK is valid. */
          if (payld /* Slot is still in the list... */
              && 0 == memcmp(pktno_net, slot->pktno_net, sizeof pktno_net)) {
            /* ... and packet numbers match. ACK is fresh. */
            if (0U == slot->payld_u8_num)
              /* ACK for EOF packet. */
              ret = 0;
            else {
              /*
                "Send_cfg->payld_u8_num" is smaller than "SSIZE_MAX", see
                definition of "SEND_PAYLD_U8_MAX", "OFS_PAYLD" being positive,
                and "args_parse()".
              */
              size_t left = send_cfg->payld_u8_num;
              ssize_t cur;

              /*
                See the comment on blocking in "udp_recv2.c" at
                "recv_pktbuf_flush()".
              */
              while (0U < left
                  && 0 < (cur = read(STDIN_FILENO, payld, left))) {
                left -= (size_t)cur;
                payld += (size_t)cur;
              }

              if (-1 == cur) {
                fprintf(stderr, "%s: read(): %s\n", pname, strerror(errno));
                ret = -1;
              }
              else
                /* Assignment. */
                if (0U == (slot->payld_u8_num
                    = send_cfg->payld_u8_num - left))
                  /* EOF on standard input. */
                  if (slot->next == slot)
                    /*
                      This is the last slot. Let it stay in the list as the
                      only slot, holding the EOF packet. Its payload length is
                      already set to zero. It still needs a new packet number,
                      otherwise the receiver won't accept it.

                      We won't read in and send any packets behind this one,
                      because (1) this is the only slot left, (2) if we get an
                      ACK for this slot, transfer is finished. Thus there is no
                      need to increment the counter.
                    */
                    pktno_hton(slot->pktno_net, send_pktbuf->pktno_next);
                  else {
                    /* This is not the last slot, unlink it. */
                    slot->payld = 0;
                    slot->prev->next = slot->next;
                    slot->next->prev = slot->prev;

                    if (send_pktbuf->slot_next == slot)
                      /*
                        We unlinked the slot that was just about being sent.
                        Step to the next slot. That slot exists, because the
                        unlinked slot was not the last one.
                      */
                      send_pktbuf->slot_next = slot->next;
                  }
                else {
                  /* Refilled slot. */
                  pktno_hton(slot->pktno_net, send_pktbuf->pktno_next);
                  if (!pktno_incr(send_pktbuf->pktno_next)) {
                    fprintf(stderr, "%s: packet number wraparound\n", pname);
                    ret = -1;
                  }
                }
            }
          }
        }
      }
  }

  return ret;
}


/*-----------------------------------------------------------------------------
  Return value:
     0 if successful.
    -1 otherwise. "Send_pktbuf" is unchanged.
*/
static int pkt_pump(struct send_pktbuf *send_pktbuf,
    const struct sockaddr_in *peer, int sock)
{
  struct send_slot *slot = send_pktbuf->slot_next;
  struct iovec fields[IDX_ALL];
  struct msghdr msg;

  fields[IDX_SSN  ].iov_base = &slot->ssn_net;
  fields[IDX_SSN  ].iov_len  = sizeof slot->ssn_net;
  fields[IDX_PKTNO].iov_base = slot->pktno_net;
  fields[IDX_PKTNO].iov_len  = sizeof slot->pktno_net;
  fields[IDX_PAYLD].iov_base = slot->payld;
  fields[IDX_PAYLD].iov_len  = slot->payld_u8_num;
  /*
    Sum of all "iov_len" fields is at most "32767U". See "PAYLD_U8_NUM",
    "SEND_PAYLD_U8_MAX" and "args_parse()".
  */

  msg.msg_name       = (void *)peer; /* Not modified by "sendmsg()". */
  msg.msg_namelen    = sizeof *peer;
  msg.msg_iov        = fields;
  msg.msg_iovlen     = sizeof fields / sizeof fields[0]; /* "Int"? */
  msg.msg_control    = 0;
  msg.msg_controllen = 0U;
  /* Ignored: "msg.msg_flags". */

  if (-1 == sendmsg(sock, &msg, 0) && -1 == check_transient(errno)) {
    fprintf(stderr, "%s: sendmsg(): %s\n", pname, strerror(errno));
    return -1;
  }

  send_pktbuf->slot_next = slot->next;
  return 0;
}


/*-----------------------------------------------------------------------------
  Return value:
     0 if successful (whole stream sent).
    -1 otherwise. "Send_pktbuf" must be uninitialized.
*/
static int snd(struct send_pktbuf *send_pktbuf, const struct sockaddr_in *peer,
    int sock, const struct send_cfg *send_cfg)
{
  /*
    Local timestamp of last valid, not necessarily fresh ACK, received from our
    peer.
  */
  struct timeval last_ack,
      prev; /* C[I-1], I >= 0 */
  int ret = 1;

  /*
    See "nat_init()" in "udp_recv2.c" for more timing-related explanations.
    Also see "timer_test.c".

    Drop any pending SIGCONT to avoid unnecessary restarting of the timer in
    the first execution of the loop.
  */
  sigcont_fetch();

  /* C[-1] := current_time */
  gettimeofday(&prev, 0);
  last_ack = prev;

  /* "Send_cfg->bps" is in "[1L, LONG_MAX / 2L + 1L]", see "args_parse()". */
  timer_init(&prev, send_cfg->bps);

  while (1 == ret) {
    /* L[I] := exec_event(I) */
    int ack = 0;

    ret = pkt_ckack(&ack, send_pktbuf, peer, sock, send_cfg);
    if (1 == ret) {
      if (-1 == pkt_pump(send_pktbuf, peer, sock))
        ret = -1;
      else {
        struct timeval now;
        int restart = 0;

        /* C[I] := current_time */
        gettimeofday(&now, 0);

        /*
          Fetch any pending SIGCONT first. See the same place of "nat_init()"
          in "udp_recv2.c" for reasoning.
        */
        if (sigcont_fetch() || -1 == tv_sub(0, &now, &prev)) {
          if (0 < fprintf(stderr, "%s: restarting timer\n", pname)) {
            last_ack = now;
            restart = 1;
          }
          else
            ret = -1;
        }

        if (1 == ret) {
          struct timeval passed;

          tv_sub(&passed, &now, &last_ack);
          if (-1 == tv_sub(0, &send_cfg->timeout, &passed)) {
            fprintf(stderr, "%s: timeout\n", pname);
            ret = -1;
          }
          else
            if (-1 == timer_adv(&now, &send_cfg->event_length, restart)) {
              fprintf(stderr, "%s: select(): %s\n", pname, strerror(errno));
              ret = -1;
            }
            else {
              if (ack)
                last_ack = now;

              /* I := I + 1 */
              prev = now;
            }
        }
      }
    }
  }

  return ret;
}


int main(int argc, char **argv)
{
  int ret = EXIT_FAILURE;

  pname = strrchr(argv[0], '/');
  pname = pname ? pname + 1 : argv[0];
  signal_init();

#ifndef __VMS
  /*
    Close "stdout", ignore return value. Don't do this with "Compaq C V6.2-003
    on OpenVMS Alpha V7.3-2", because if "stdin" is redirected from a file (not
    a terminal), then the following sequence leads, for some reason, to
    "%SYSTEM-F-ACCVIO, access violation, reason mask=00, virtual
    address=000000000000001C, PC=FFFFFFFF809CE448, PS=0000001B":

    (1) fclose(stdout);
    (2) STDOUT_FILENO == socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    (3) read(STDIN_FILENO, &x, 1);

    (1) was executed here, (2) was executed in "sock_init()", (3) triggered the
    crash in "pkt_ckack()".
  */
  (void)fclose(stdout);
#endif

  if (-1 == check_time_t())
    fprintf(stderr, "%s: time_t is not an integral type, unsupported"
        " platform\n", pname);
  else {
    struct send_cfg send_cfg;

    ret = args_parse(&send_cfg, argc, argv);
    if (-1 == ret) {
      struct send_pktbuf send_pktbuf;

      ret = EXIT_FAILURE;
      if (-1 != send_pktbuf_init(&send_pktbuf, &send_cfg)) {
        int sock = sock_init(send_cfg.wait_mode ? &send_cfg.addr : 0);

        if (-1 != sock) {
          struct sockaddr_in initiator;
          const struct sockaddr_in *peer = 0;

          if (send_cfg.wait_mode) {
            /*
              The "sock" argument to "pkt_wait()" is OK: see the description of
              "sock_init()". The "timeout" argument to "pkt_wait()" is OK: see
              "args_parse()" and "MAXCDIFFL".
            */
            if (-1 != pkt_wait(sock, &send_cfg.timeout)) {
              size_t address_len = sizeof initiator;
              char unsigned vms_dummy; /* See definition in "udp_recv2.c". */

              /* Drop first packet, take address. Don't tolerate errors. */
              if (-1 == recvfrom(sock, &vms_dummy, 0, 0,
                  (struct sockaddr *)&initiator, &address_len))
                fprintf(stderr, "%s: recvfrom(): %s\n", pname,
                    strerror(errno));
              else
                peer = &initiator;
            }
          }
          else
            peer = &send_cfg.addr;

          if (peer && -1 != snd(&send_pktbuf, peer, sock, &send_cfg)) {
#ifndef __VMS
            /* Actually, this isn't necessary even under UNIX. */
            if (EOF == fclose(stdin))
              fprintf(stderr, "%s: fclose(): %s\n", pname, strerror(errno));
            else
#endif
              ret = EXIT_SUCCESS;
          }
          close(sock);
        }
        send_pktbuf_uninit(&send_pktbuf);
      }
    }
  }

  return ret;
}
