/* $Id: uwd_progress.c,v 1.10 2005/06/04 21:50:12 lacos Exp $ */

#define _XOPEN_SOURCE
#define _XOPEN_SOURCE_EXTENDED 1 /* enable X/Open UNIX Extension */

#include <sys/time.h> /* gettimeofday() */
#include <float.h>    /* FLT_RADIX */
#include <errno.h>    /* errno */
#include <string.h>   /* strerror() */
#include <time.h>     /* time() */
#include <stdio.h>    /* fprintf() */
#include <stdlib.h>   /* EXIT_FAILURE */

#include "uwd.h"      /* UWD_U32_NUM */


static const u32
    uwd_million[UWD_U32_NUM] = { 1000000LU },
    uwd_31days[UWD_U32_NUM] = { 0x9d14a000LU, 0x26fLU }; /* usecs */


/*-----------------------------------------------------------------------------
  Convert the timestamp in "tp" into usecs, storing it into "result".

  The type "time_t" must be an integral type. "Tp" must be valid, normalized,
  and non-negative. If any of these constraints is violated, the behaviour is
  undefined.

  Return value:
    "Result" if conversion is successful.
        NULL otherwise (range error). "Result" is indeterminate.
*/
static u32 *tv2uwd(u32 result[UWD_U32_NUM], const struct timeval *tp)
{
  u32 *tmp_r = result;
  const u32 *end_r = result + UWD_U32_NUM;
  time_t tv_sec = tp->tv_sec;

  while (tmp_r != end_r) {
    *tmp_r++ = tv_sec;
    if (sizeof(time_t) * (unsigned)CHAR_BIT > U32_BIT_NUM)
      tv_sec >>= U32_BIT_NUM;
    else
      tv_sec = 0;
  }

  if ((time_t)0 == tv_sec && uwd_mul(result, result, uwd_million)) {
    u32 tv_usec[UWD_U32_NUM];

    uwd_u32(tv_usec, tp->tv_usec);
    if (!uwd_add(result, result, tv_usec))
      return result;
  }

  return 0;
}


/*-----------------------------------------------------------------------------
  Convert the usecs stored in "a" into "tp".

  Since "select()" supports a timeout of at least 31 days, "struct timeval" can
  represent at least 31 days. Thus, for portability, this function enforces the
  inclusive upper bound of "uwd_31days" for "a".

  Return value:
    "Result" if the conversion is successful. "Result" is normalized.
        NULL otherwise (range error). "Result" is unchanged.
*/
static struct timeval *uwd2tv(struct timeval *result, const u32 a[U32_BIT_NUM])
{
  if (-1 != uwd_cmp(uwd_31days, a)) {
    u32 tv_sec[UWD_U32_NUM],
        tv_usec[UWD_U32_NUM];

    uwd_div(tv_sec, tv_usec, a, uwd_million);
    result->tv_sec = tv_sec[0]; /* valid, see above */
    result->tv_usec = tv_usec[0];
    return result;
  }

  return 0;
}


/*-----------------------------------------------------------------------------
  Check whether "time_t" is an integral type or not.

  Return value:
    1 if "time_t" is integral.
    0 otherwise.
*/
static int integral_time_t(void)
{
  return (time_t)0 == (time_t)(1.0 / FLT_RADIX);
}


/*-----------------------------------------------------------------------------
  Structure for linear progress shaping.
*/
struct progress {
  u32 num_bytes[UWD_U32_NUM], /* work done so far */
      base[UWD_U32_NUM],      /* timestamp of initialization, usecs */
      bps[UWD_U32_NUM];       /* wanted average speed */
};


/*-----------------------------------------------------------------------------
  Initialize "progress" from the current time and "str_bps". "Num_bytes" is set
  to zero.

  Return value:
    "Progress" if successful.
          NULL if failed to convert the current time into "base", or "str_bps"
               into "bps". "Progress" is indeterminate.
*/
static struct progress *progress_init(struct progress *progress,
    const char *str_bps)
{
  struct timeval tv;

  gettimeofday(&tv, 0);
  if (tv2uwd(progress->base, &tv) && str2uwd(progress->bps, str_bps)) {
    uwd_cpy(progress->num_bytes, uwd_zero);
    return progress;
  }

  return 0;
}


/*-----------------------------------------------------------------------------
  Advance "adv_bytes".

  Sleep for the correct amount of time, required by the average speed of "bps".
  Error messages are logged.

  Return value:
    "Progress" if successful.
          NULL otherwise. Contents of "progress" are indeterminate.
*/
static struct progress *progress_adv(struct progress *progress,
    const u32 adv_bytes[UWD_U32_NUM])
{
  struct progress *ret = 0;
  u32 *num_bytes = progress->num_bytes;

  if (-1 == uwd_add(num_bytes, num_bytes, adv_bytes))
    fprintf(stderr, "uwd overflow\n");
  else {
    u32 to_elapse[UWD_U32_NUM];

    if (!uwd_mul(to_elapse, num_bytes, uwd_million))
      fprintf(stderr, "uwd overflow\n");
    else  {
      u32 dummy_rem[UWD_U32_NUM];

      if (-1 == uwd_div(to_elapse, dummy_rem, to_elapse, progress->bps))
        fprintf(stderr, "uwd div0\n");
      else {
        struct timeval tv_now;
        u32 now[UWD_U32_NUM];

        gettimeofday(&tv_now, 0);
        if (!tv2uwd(now, &tv_now))
          fprintf(stderr, "struct timeval -> uwd overflow\n");
        else {
          u32 elapsed[UWD_U32_NUM];

          if (-1 == uwd_sub(elapsed, now, progress->base))
            fprintf(stderr, "time skew");
          else {
            u32 to_sleep[UWD_U32_NUM];

            ret = progress;
            if (1 == uwd_sub(to_sleep, to_elapse, elapsed)) {
              struct timeval tv_to_sleep;

              ret = 0;

              if (!uwd2tv(&tv_to_sleep, to_sleep))
                fprintf(stderr, "uwd -> limited struct timeval overflow\n");
              else
                if (-1 == select(0, 0, 0, 0, &tv_to_sleep))
                  fprintf(stderr, "select(): %s\n", strerror(errno));
                else
                  ret = progress;
            }
          }
        }
      }
    }
  }

  return ret;
}


int main(int argc, char **argv)
{
  if (!integral_time_t())
    fprintf(stderr, "time_t is not an integral type, unsupported platform\n");
  else {
    if (2 != argc)
      fprintf(stderr, "usage: %s BPS\n", argv[0]);
    else {
      u8 buf[2048];
      struct progress pr;

      memset(buf, '1', sizeof buf);
      srand(time(0));

      if (!progress_init(&pr, argv[1]))
        fprintf(stderr, "failed to initialize progress\n");
      else {
        int cont = 1;
        u32 adv_bytes[UWD_U32_NUM] = { 0 };

        while (cont && pr.num_bytes[1] < 1U) {
          adv_bytes[0] = 1024U + (unsigned)rand() % 1025U;

          cont = 0;

          if (adv_bytes[0] != fwrite(buf, 1, adv_bytes[0], stdout)
              || EOF == fflush(stdout))
            fprintf(stderr, "fwrite(): %s\n", strerror(errno));
          else
            cont = (&pr == progress_adv(&pr, adv_bytes));
        }
      }
    }
  }

  return EXIT_FAILURE;
}
